from dataclasses import dataclass
from typing import List, Optional

import pandas as pd
from dataclasses_json import dataclass_json, LetterCase


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class GeneratedResponse:
    prompt: str
    response: str
    raw_prompt: Optional[str] = None
    sample_index: Optional[int] = None
    sample_data: Optional[str] = None
    context: Optional[str] = None
    model_name: Optional[str] = None
    finish_reason: Optional[str] = None
    generated_tokens: Optional[int] = None

    @staticmethod
    def to_responses(df: pd.DataFrame) -> List["GeneratedResponse"]:
        return [GeneratedResponse(**row.to_dict()) for _, row in df.iterrows()]



import enum
from abc import ABCMeta


class ConnectionType(enum.Enum):
    FILE = "file"
    PUBLIC_DATASETS = "public_datasets"
    ADLS = "adls"
    GCS = "gcs"
    S3 = "s3"
    MYSQL = "mysql"
    POSTGRESQL = "postgresql"
    SNOWFLAKE = "snowflake"
    REDSHIFT = "redshift"
    BIGQUERY = "bigquery"
    DATABRICKS = "databricks"


class ConnectionProperties(metaclass=ABCMeta):
    def __init__(self):
        pass

    def to_dict(self):
        return self.__dict__


class DatabaseProperties(ConnectionProperties):
    def __init__(self, address: str, database: str, username: str, password: str):
        self.address = address
        self.database = database
        self.username = username
        self.password = password


class SnowflakeProperties(ConnectionProperties):
    def __init__(
        self,
        user: str,
        password: str,
        account: str,
        database: str,
        warehouse: str,
        schema: str,
    ):
        self.user = user
        self.password = password
        self.account = account
        self.database = database
        self.warehouse = warehouse
        self.schema = schema


class S3Properties(ConnectionProperties):
    def __init__(self, key: str, secret: str):
        self.key = key
        self.secret = secret
        self.protocol = "s3"


class BigQueryProperties(ConnectionProperties):
    def __init__(self, project: str, dataset: str, token: str):
        self.project = project
        self.dataset = dataset
        self.token = token


class GCSProperties(ConnectionProperties):
    def __init__(self, token: str):
        self.protocol = "gcs"
        self.token = token


class ADLSProperties(ConnectionProperties):
    def to_dict(self):
        return {
            "accountKey": self.account_key,
            "connectionString": self.connection_string,
            "protocol": self.protocol,
        }

    def __init__(self, account_key: str, connection_string: str):
        self.account_key = account_key
        self.connection_string = connection_string
        self.protocol = "abfs"


class DatabricksProperties(ConnectionProperties):
    def __init__(self, access_token: str, server_host_name: str, http_path: str, schema_name: str = None):
        self.access_token = access_token
        self.server_host_name = server_host_name
        self.http_path = http_path
        self.schema_name = schema_name

    def to_dict(self):
        return {
            "serverHostname": self.server_host_name,
            "httpPath": self.http_path,
            "token": self.access_token,
            "schema": self.schema_name,
        }



from typing import Optional

import pandas as pd
import typer
from rich import box
from rich.progress import Progress, SpinnerColumn, TextColumn
from rich.table import Table

from predibase.cli_commands.utils import get_client, get_console

app = typer.Typer()


@app.command(help="Query a Large Language Model (LLM)")
def llm(
    template: str = typer.Option(
        None,
        "--template",
        "-t",
        prompt="Template input",
        prompt_required=True,
        help="Prompt template input text",
    ),
    model_name: str = typer.Option(
        None,
        "--model-name",
        "-m",
        prompt="Model to query",
        prompt_required=True,
        help="Name of the model",
    ),
    index_name: Optional[str] = typer.Option(
        None,
        "--index-name",
        "-i",
        prompt_required=False,
        help="Optional dataset to index",
    ),
    dataset_name: Optional[str] = typer.Option(
        None,
        "--dataset-name",
        "-d",
        prompt_required=False,
        help="Optional dataset for batch inference",
    ),
    limit: Optional[int] = typer.Option(None, "--limit", "-l", prompt_required=False, help="Optional results limit"),
):
    client = get_client()

    with Progress(
        SpinnerColumn(),
        TextColumn("[progress.description]{task.description}"),
        transient=True,
    ) as progress:
        progress.add_task(description="Querying LLM...", total=None)

        df = client.prompt(
            template,
            model_name,
            index=index_name,
            dataset=dataset_name,
            limit=limit,
            return_df=True,
        )

    table = Table(show_header=True, header_style="bold magenta")

    # Modify the table instance to have the data from the DataFrame
    table = df_to_table(df, table)

    # Update the style of the table
    table.row_styles = ["none", "dim"]
    table.box = box.SIMPLE_HEAD

    get_console().print(table)


def df_to_table(
    pandas_dataframe: pd.DataFrame,
    rich_table: Table,
    show_index: bool = True,
    index_name: Optional[str] = None,
) -> Table:
    """Convert a pandas.DataFrame obj into a rich.Table obj.

    Args:
        pandas_dataframe (DataFrame): A Pandas DataFrame to be converted to a rich Table.
        rich_table (Table): A rich Table that should be populated by the DataFrame values.
        show_index (bool): Add a column with a row count to the table. Defaults to True.
        index_name (str, optional): The column name to give to the index column. Defaults to None, showing no value.
    Returns:
        Table: The rich Table instance passed, populated with the DataFrame values.
    """

    if show_index:
        index_name = str(index_name) if index_name else ""
        rich_table.add_column(index_name)

    for column in pandas_dataframe.columns:
        rich_table.add_column(str(column))

    for index, value_list in enumerate(pandas_dataframe.values.tolist()):
        row = [str(index)] if show_index else []
        row += [str(x) for x in value_list]
        rich_table.add_row(*row)

    return rich_table


if __name__ == "__main__":
    app()



import json
import os
from pathlib import Path
from typing import Any, Dict

import typer
from rich import print

APP_NAME = "predibase-sdk-app"
LOCAL_SETTINGS_NAME = ".predibase.json"

app = typer.Typer()


@app.command(help="Show global settings")
def show_global():
    settings = load_global_settings()
    print(settings)


@app.command(help="Show local settings")
def show_local():
    settings = load_local_settings()
    print(settings)


@app.command(help="Show local and global settings")
def show_all():
    settings = load_settings()
    print(settings)


@app.command(help="Set default repository name")
def set_repo(repository_name: str = typer.Argument(..., help="The model repository name")):
    _set_setting("repo", repository_name)


@app.command(help="Set default engine name")
def set_engine(engine_name: str = typer.Argument(..., help="The engine name")):
    _set_setting("engine", engine_name)


@app.command(help="Set api token")
def set_api_token(token: str = typer.Argument(..., help="The api token")):
    _set_setting("token", token)


@app.command(help="Set default endpoint url")
def set_endpoint(endpoint: str = typer.Argument(..., help="The endpoint url")):
    _set_setting("endpoint", endpoint)


def _set_setting(setting: str, value: Any):
    settings = load_global_settings()
    settings[setting] = value
    save_global_settings(settings)


def get_global_settings_path() -> Path:
    app_dir = typer.get_app_dir(APP_NAME)
    return Path(app_dir) / "settings.json"


def save_global_settings(settings: Dict[str, Any]):
    settings_path = get_global_settings_path()
    os.makedirs(settings_path.parent, exist_ok=True)
    with open(settings_path, "w") as f:
        json.dump(settings, f)


def load_global_settings() -> Dict[str, Any]:
    settings_path = get_global_settings_path()
    if settings_path.is_file():
        with open(settings_path) as f:
            return json.load(f)
    return {}


def save_local_settings(settings: Dict[str, Any]):
    settings_path = Path(LOCAL_SETTINGS_NAME)
    os.makedirs(settings_path.parent, exist_ok=True)
    with open(settings_path, "w") as f:
        json.dump(settings, f)


def load_local_settings() -> Dict[str, Any]:
    settings_path = Path(LOCAL_SETTINGS_NAME)
    if settings_path.is_file():
        with open(settings_path) as f:
            return json.load(f)
    return {}


def load_settings() -> Dict[str, Any]:
    local_settings = load_local_settings()
    global_settings = load_global_settings()
    return {**global_settings, **local_settings}



from abc import abstractmethod
from typing import Dict, List, Optional, Union

import pandas as pd

from predibase.pql.api import Session
from predibase.resource.connection import Connection
from predibase.resource.dataset import Dataset
from predibase.resource.llm import GeneratedResponse


class LlmMixin:
    session: Session

    def prompt(
        self,
        templates: Union[str, List[str]],
        model_name: Union[str, List[str]],
        index_name: Optional[Union[str, Dataset]] = None,
        dataset_name: Optional[Union[str, Dataset]] = None,
        limit: Optional[int] = None,
        options: Optional[Dict[str, float]] = None,
        return_df: bool = False,
    ) -> Union[List[GeneratedResponse], pd.DataFrame]:
        options_str = ", ".join(f"{k}={v}" for k, v in options.items()) if options else ""
        options_clause = f"WITH OPTIONS ({options_str}) " if options_str else ""

        models = [model_name] if isinstance(model_name, str) else model_name
        models_str = ", ".join(f'"{m}"' for m in models)
        model_clause = f"USING {models_str} " if models_str else ""

        index = self.get_dataset(index_name) if isinstance(index_name, str) else index_name
        index_clause = f'OVER "{index.connection.name}"."{index.name}" ' if index else ""

        dataset = self.get_dataset(dataset_name) if isinstance(dataset_name, str) else dataset_name
        limit_clause = f" LIMIT {limit}" if limit else ""
        given_clause = (
            f'GIVEN select * from "{dataset.connection.name}"."{dataset.name}"{limit_clause}' if dataset else ""
        )

        templates = [templates] if isinstance(templates, str) else templates
        templates_str = ", ".join(f"'{t}'" for t in templates)
        query_str = f"PROMPT {templates_str} {options_clause}{model_clause}{index_clause}{given_clause};"

        conn_id = None
        if isinstance(dataset, Dataset):
            conn_id = dataset.connection_id
        elif isinstance(index, Dataset):
            conn_id = index.connection_id
        else:
            conn_id = self.list_connections()[0].id

        df = self.session.execute(
            query_str,
            connection_id=conn_id,
        )
        return GeneratedResponse.to_responses(df) if not return_df else df

    def deploy_llm(self, deployment_name: str, model_name: str, engine_template: Optional[str] = None):
        self.session.post_json(
            "/llms",
            json={
                "name": deployment_name,
                "modelName": model_name,
                "engineTemplate": engine_template,
            },
        )

    def delete_llm(self, deployment_name: str):
        self.session.delete_json(f"/llms/{deployment_name}")

    @abstractmethod
    def get_dataset(self) -> Dataset:
        pass

    @abstractmethod
    def list_connections(self) -> List[Connection]:
        pass



from dataclasses import dataclass
from typing import Any, Dict, Optional

import yaml
from rich.console import Console

from predibase import PredibaseClient
from predibase.pql.api import Session
from predibase.resource.engine import Engine
from predibase.resource.model import ModelRepo


@dataclass
class Defaults:
    repo: Optional[str] = None
    engine: Optional[str] = None
    session: Optional[Session] = None


console = Console()
defaults: Defaults = Defaults()


def get_console() -> Console:
    return console


def get_client() -> PredibaseClient:
    return PredibaseClient(session=defaults.session)


def sanitize_engine_name(name: str) -> str:
    return name.replace("/", "-")


def get_repo(repo_name: Optional[str] = None) -> ModelRepo:
    repo_name = repo_name or defaults.repo
    if repo_name is None:
        raise ValueError("Repo name is required")
    return get_client().get_model_repo(name=repo_name)


def get_engine(engine_name: Optional[str] = None) -> Engine:
    engine_name = engine_name or defaults.engine
    if engine_name is None:
        raise ValueError("Engine name is required")
    return get_client().get_engine(name=engine_name)


def set_defaults_from_settings(settings: Dict[str, Any]):
    global defaults
    defaults = Defaults(
        repo=settings.get("repo"),
        engine=settings.get("engine"),
        session=Session(token=settings.get("token"), url=settings.get("endpoint")),
    )


def load_yaml(fname: str) -> Dict[str, Any]:
    with open(fname) as f:
        return yaml.safe_load(f)



import logging
from dataclasses import dataclass, field
from enum import Enum
from typing import Optional

from dataclasses_json import config, dataclass_json, LetterCase

from predibase.pql.api import Session


class DefaultEngineType(Enum):
    QUERY = "query"
    TRAIN = "train"


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class Engine:
    session: Session
    _id: int = field(metadata=config(field_name="id"))
    _type: str = field(metadata=config(field_name="type"))
    _name: str = field(metadata=config(field_name="name"))
    _auto_suspend_seconds: Optional[int] = field(metadata=config(field_name="autoSuspendSeconds"))
    _template_id: int = field(metadata=config(field_name="templateID"))
    _environment_id: int = field(metadata=config(field_name="environmentID"))
    _service_type: str = field(metadata=config(field_name="serviceType"))
    _auto_resume: Optional[bool] = field(metadata=config(field_name="autoResume"), default=False)
    # _user: dict = field(metadata=config(field_name="user"))

    def __repr__(self):
        return (
            f"Engine(id={self.id}, name={self.name}, status={self.status}, type={self.type}, "
            f"service_type={self.service_type}, template_id={self.template_id}, environment_id={self.environment_id}, "
            f"auto_suspend_seconds={self.auto_suspend_seconds}, auto_resume={self.auto_resume})"
        )

    @property
    def id(self):
        return self._id

    @property
    def type(self):
        return self._type

    @property
    def service_type(self):
        return self._service_type

    @property
    def name(self):
        return self._name

    # @property
    # def author(self):
    #     if
    #     return self._user["username"]

    def set_name(self, value):
        self._name = value
        self._update()

    @property
    def auto_suspend_seconds(self):
        return self._auto_suspend_seconds

    def set_auto_suspend_seconds(self, value):
        self._auto_suspend_seconds = value
        self._update()

    @property
    def auto_resume(self):
        return self._auto_resume

    def set_auto_resume(self, value):
        self._auto_resume = value
        self._update()

    @property
    def instance_type_id(self):
        return self._instance_type_id

    @property
    def template_id(self):
        return self._template_id

    @property
    def environment_id(self):
        return self._environment_id

    def set_template_id(self, value):
        self._template_id = value
        self._update()

    def _get(self):
        return self.session.get_json(f"/engines/{self.id}")

    @property
    def status(self):
        return self._get()["engineStatus"]

    def __dict__(self):
        return {
            "id": self.id,
            "type": self.type,
            "name": self.name,
            "autoSuspendSeconds": self.auto_suspend_seconds,
            "autoResume": self.auto_resume,
            "templateId": self.template_id,
            "environmentId": self.environment_id,
            "serviceType": self.service_type,
        }

    def _update(self):
        resp = self.session.put_json(f"/engines/{self.id}", self.__dict__())
        return resp

    def start(self):
        if not self.session.is_plan_expired():
            try:
                self.session.post_json("/engines/start", self.__dict__())
                resp = self.session.get_json_until(
                    f"/engines/{self.id}",
                    lambda resp: resp["engineStatus"] == "active",
                    lambda resp: "Engine failed to start with status ERRORED"
                    if resp["engineStatus"] == "errored"
                    else None,
                )
                return resp
            except RuntimeError:
                logging.info("Engine is already active")
        else:
            raise PermissionError(
                "Engines are locked for expired plans. Contact us to upgrade.",
            )

    def stop(self):
        try:
            self.session.post_json("/engines/stop", self.__dict__())
            resp = self.session.get_json_until(
                f"/engines/{self.id}",
                lambda resp: resp["engineStatus"] == "suspended",
                lambda resp: "Engine failed to start with status ERRORED"
                if resp["engineStatus"] == "errored"
                else None,
            )
            return resp
        except RuntimeError:
            logging.info("Engine is already suspended")



from dataclasses import dataclass, field
from typing import Optional, Union

from dataclasses_json import config, dataclass_json, LetterCase


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class Subscription:
    _id: int = field(metadata=config(field_name="id"))
    _tier: str = field(metadata=config(field_name="tier"))
    _is_trial: bool = field(metadata=config(field_name="isTrial"))
    _expiration: Union[str, None] = field(metadata=config(field_name="expiration"))
    _days_remaining_in_plan: Union[int, None] = field(metadata=config(field_name="daysRemainingInPlan"))
    _active: bool = field(metadata=config(field_name="active"))

    def __repr__(self):
        return (
            f"Subscription(id={self._id}, tier={self._tier}, is_trial={self._is_trial}, "
            f"expiration={self._expiration}, days_remaining_in_plan={self._days_remaining_in_plan}, "
            f"active={self._active})"
        )

    @property
    def id(self):
        return self._id

    @property
    def tier(self):
        return self._tier

    @property
    def is_trial(self):
        return self._is_trial

    @property
    def expiration(self):
        return self._expiration

    @property
    def days_remaining_in_plan(self):
        return self._days_remaining_in_plan

    @property
    def active(self):
        return self._active


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class Tenant:
    _name: str = field(metadata=config(field_name="name"))
    _short_code: str = field(metadata=config(field_name="shortCode"))
    _subscription: Subscription = field(metadata=config(field_name="subscription"))

    def __repr__(self):
        return f"Tenant(name={self._name}, short_code={self._short_code}, subscription={self._subscription})"

    @property
    def name(self):
        return self._name

    @property
    def short_code(self):
        return self._short_code

    @property
    def subscription(self):
        return self._subscription


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class User:
    _id: int = field(metadata=config(field_name="id"))
    _username: str = field(metadata=config(field_name="username"))
    _uuid: str = field(metadata=config(field_name="uuid"))
    _tenant: Tenant = field(metadata=config(field_name="tenant"))
    _comment: Optional[str] = field(metadata=config(field_name="comment"), default=None)

    def __repr__(self):
        # Only return the username and uuid as relevant information
        return f"User(id={self._uuid}, username={self._username})"

    # Users should use the kratos UUID for RBAC
    @property
    def id(self):
        return self._uuid

    @property
    def user_id(self):
        return self._user_id

    @property
    def username(self):
        return self._username

    @property
    def tenant(self) -> Tenant:
        return self._tenant

    # def set_user_name(self, name: str):

    # def _update(self):
    #     resp = self.session.post_json("/users/update", self.__dict__())



from typing import Optional, Union

from predibase.connection_mixin import ConnectionMixin
from predibase.dataset_mixin import DatasetMixin
from predibase.deployment_mixin import DeploymentMixin
from predibase.engine_mixin import EngineMixin
from predibase.llm_mixin import LlmMixin
from predibase.model_mixin import ModelMixin
from predibase.permission_mixin import PermissionMixin
from predibase.pql import get_session
from predibase.pql.api import Session
from predibase.query_mixin import QueryMixin


class PredibaseClient(
    QueryMixin,
    ConnectionMixin,
    DatasetMixin,
    EngineMixin,
    ModelMixin,
    DeploymentMixin,
    PermissionMixin,
    LlmMixin,
):
    def __init__(self, session: Optional[Session] = None, gateway: str = None, token: str = None):
        self._session = session or get_session(gateway, token)

    @property
    def session(self) -> Session:
        return self._session

    def set_connection(self, connection: Union[str, int]):
        """This method sets the default connection to use for any method requiring a connection argument.

        :param connection: The desired connection name or id
        """
        self.session.set_connection(connection)



import re
from typing import Optional

import typer

from predibase.cli_commands.utils import get_client, get_console

app = typer.Typer()


# TODO(travis): this should be dynamic based on model type and pushed to the backend
DEFAULT_TEMPLATE = "llm-gpu-small"


@app.command(help="Create a Large Language Model (LLM) deployment")
def llm(
    deployment_name: str = typer.Option(
        None,
        "--deployment-name",
        "-d",
        prompt="Deployment name",
        prompt_required=True,
        help="Name of the deployment",
    ),
    model_name: str = typer.Option(
        None,
        "--model-name",
        "-m",
        prompt="Model to deploy",
        prompt_required=True,
        help="Name of the model",
    ),
    engine_template: Optional[str] = typer.Option(
        DEFAULT_TEMPLATE,
        "--engine-template",
        "-e",
        prompt="Engine name",
        prompt_required=False,
        help="Optional engine template to provision for hosting the model",
    ),
    # auto_suspend_secs: Optional[int] = 3600,  # TODO: add ability to auto suspend
):
    # raise ValueError if name is not lower case alphanumeric characters or '-'
    if re.match(r"^[a-z0-9-]+$", deployment_name) is None:
        raise ValueError("name must be lower case alphanumeric characters or '-'")

    client = get_client()

    get_console().print("Deploying an LLM with the following parameters:")
    get_console().print("\tdeployment_name:", deployment_name)
    get_console().print("\tmodel_name:", model_name)
    get_console().print("\tengine_template:", engine_template)

    client.deploy_llm(deployment_name, model_name, engine_template=engine_template)
    get_console().print("Deploy request sent.")


if __name__ == "__main__":
    app()



from typing import Any, Dict, List, Optional, Union

import pandas as pd

from predibase.connection_mixin import ConnectionMixin
from predibase.resource.query import Query


class QueryMixin(ConnectionMixin):
    def query_history(
        self,
        offset: int = 0,
        limit: int = 10,
        raw_query: Optional[str] = None,
        df: bool = False,
    ) -> Union[pd.DataFrame, List[Query]]:
        """
        :param offset
        :param limit
        :param raw_query: query text to search on
        :param df: Whether to return pandas dataframe or list of objects
        :return: Pandas dataframe or list of modelRepos
        """
        endpoint = f"/queries?offset={offset}&limit={limit}"
        if raw_query:
            endpoint += f"&searchKeys[]=rawQuery&searchVals[]={raw_query}"
        resp = self.session.get_json(endpoint)
        queries = [x for x in resp["queryHistory"]]
        if df:
            return pd.DataFrame(queries)
        return [Query.from_dict({"session": self.session, **x}) for x in queries]

    # def get_queries_by_text(self, raw_text: str) -> List[Query]:
    #     resp = self.get(f"/queries?searchKeys[]=rawText,searchVals[]={raw_text}?limit=999999999")
    #     return [Query({"session": self.session, **q}) for q in resp["queryHistory"]]

    def set_query_connection(self, connection_name: str):
        conn = self.get_connection(connection_name)
        self.session.connection_id = conn.id

    def execute_query(
        self,
        query: str,
        params: Optional[Dict[str, Any]] = None,
        connection_name: Optional[str] = None,
    ) -> pd.DataFrame:
        if not self.session.is_plan_expired():
            if connection_name:
                self.set_query_connection(connection_name)
            return self.session.execute(query, params)
        else:
            raise PermissionError(
                "Queries are locked for expired plans. Contact us to upgrade.",
            )
